# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""
The SASS language brings countless amazing features to CSS.

Besides being used in web development, CSS is also the way to stylize Qt-based
desktop applications. However, Qt's CSS has a few variations that prevent the
direct use of SASS compiler.

The purpose of qtsass is to fill the gap between SASS and Qt-CSS by handling
those variations.
"""

# yapf: disable

from __future__ import absolute_import

# Standard library imports
import logging

# Local imports
from qtsass.api import (
    compile,
    compile_dirname,
    compile_filename,
    enable_logging,
    watch,
)


# yapf: enable

# Setup Logging
logging.getLogger(__name__).addHandler(logging.NullHandler())
enable_logging()

# Constants
__version__ = '0.3.1'


def _to_version_info(version):
    """Convert a version string to a number and string tuple."""
    parts = []
    for part in version.split('.'):
        try:
            part = int(part)
        except ValueError:
            pass

        parts.append(part)

    return tuple(parts)


VERSION_INFO = _to_version_info(__version__)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""The filesystem watcher api."""

# yapf: disable

from __future__ import absolute_import

# Standard library imports
import functools
import logging
import time


_log = logging.getLogger(__name__)


def retry(n, interval=0.1):
    """Retry a function or method n times before raising an exception.

    :param n: Number of times to retry
    :param interval: Time to sleep before attempts
    """
    def decorate(fn):
        @functools.wraps(fn)
        def attempt(*args, **kwargs):
            attempts = 0
            while True:
                try:
                    return fn(*args, **kwargs)
                except Exception:
                    attempts += 1
                    if n <= attempts:
                        raise
                    time.sleep(interval)

        return attempt

    return decorate

# yapf: enable


class Watcher(object):
    """Watcher base class.

    Watchers monitor a file or directory and call the on_change method when a
    change occurs. The on_change method should trigger the compiler function
    passed in during construction and dispatch the result to all connected
    callbacks.

    Watcher implementations must inherit from this base class. Subclasses
    should perform any setup required in the setup method, rather than
    overriding __init__.
    """

    def __init__(self, watch_dir, compiler, args=None, kwargs=None):
        """Store initialization values and call Watcher.setup."""
        self._watch_dir = watch_dir
        self._compiler = compiler
        self._args = args or ()
        self._kwargs = kwargs or {}
        self._callbacks = set()
        self._log = _log
        self.setup()

    def setup(self):
        """Perform any setup required here.

        Rather than implement __init__, subclasses can perform any setup in
        this method.
        """
        return NotImplemented

    def start(self):
        """Start this Watcher."""
        return NotImplemented

    def stop(self):
        """Stop this Watcher."""
        return NotImplemented

    def join(self):
        """Wait for this Watcher to finish."""
        return NotImplemented

    @retry(5)
    def compile(self):
        """Call the Watcher's compiler."""
        self._log.debug(
            'Compiling sass...%s(*%s, **%s)',
            self._compiler,
            self._args,
            self._kwargs,
        )
        return self._compiler(*self._args, **self._kwargs)

    def compile_and_dispatch(self):
        """Compile and dispatch the resulting css to connected callbacks."""
        self._log.debug('Compiling and dispatching....')

        try:
            css = self.compile()
        except Exception:
            self._log.exception('Failed to compile...')
            return

        self.dispatch(css)

    def dispatch(self, css):
        """Dispatch css to connected callbacks."""
        self._log.debug('Dispatching callbacks...')
        for callback in self._callbacks:
            callback(css)

    def on_change(self):
        """Call when a change is detected.

        Subclasses must call this method when they detect a change. Subclasses
        may also override this method in order to manually compile and dispatch
        callbacks. For example, a Qt implementation may use signals and slots
        to ensure that compiling and executing callbacks happens in the main
        GUI thread.
        """
        self._log.debug('Change detected...')
        self.compile_and_dispatch()

    def connect(self, fn):
        """Connect a callback to this Watcher.

        All callbacks are called when a change is detected. Callbacks are
        passed the compiled css.
        """
        self._log.debug('Connecting callback: %s', fn)
        self._callbacks.add(fn)

    def disconnect(self, fn):
        """Disconnect a callback from this Watcher."""
        self._log.debug('Disconnecting callback: %s', fn)
        self._callbacks.discard(fn)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Contains the fallback implementation of the Watcher api."""

# yapf: disable

from __future__ import absolute_import, print_function

# Standard library imports
import os

# Local imports
from qtsass.importers import norm_path


# yapf: enable


def take(dir_or_file, depth=3):
    """Return a dict mapping files and folders to their mtimes."""
    if os.path.isfile(dir_or_file):
        path = norm_path(dir_or_file)
        return {path: os.path.getmtime(path)}

    if not os.path.isdir(dir_or_file):
        return {}

    snapshot = {}
    base_depth = len(norm_path(dir_or_file).split('/'))

    for root, subdirs, files in os.walk(dir_or_file):

        path = norm_path(root)
        if len(path.split('/')) - base_depth == depth:
            subdirs[:] = []

        snapshot[path] = os.path.getmtime(path)
        for f in files:
            path = norm_path(root, f)
            snapshot[path] = os.path.getmtime(path)

    return snapshot


def diff(prev_snapshot, next_snapshot):
    """Return a dict containing changes between two snapshots."""
    changes = {}
    for path in set(prev_snapshot.keys()) | set(next_snapshot.keys()):
        if path in prev_snapshot and path not in next_snapshot:
            changes[path] = 'Deleted'
        elif path not in prev_snapshot and path in next_snapshot:
            changes[path] = 'Created'
        else:
            prev_mtime = prev_snapshot[path]
            next_mtime = next_snapshot[path]
            if next_mtime > prev_mtime:
                changes[path] = 'Changed'
    return changes



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Contains the fallback implementation of the Watcher api."""

# yapf: disable

from __future__ import absolute_import, print_function

# Standard library imports
import atexit
import threading

# Local imports
from qtsass.watchers import snapshots
from qtsass.watchers.api import Watcher


# yapf: enable


class PollingThread(threading.Thread):
    """A thread that fires a callback at an interval."""

    def __init__(self, callback, interval):
        """Initialize the thread.

        :param callback: Callback function to repeat.
        :param interval: Number of seconds to sleep between calls.
        """
        super(PollingThread, self).__init__()
        self.daemon = True
        self.callback = callback
        self.interval = interval
        self._shutdown = threading.Event()
        self._stopped = threading.Event()
        self._started = threading.Event()
        atexit.register(self.stop)

    @property
    def started(self):
        """Check if the thread has started."""
        return self._started.is_set()

    @property
    def stopped(self):
        """Check if the thread has stopped."""
        return self._stopped.is_set()

    @property
    def shutdown(self):
        """Check if the thread has shutdown."""
        return self._shutdown.is_set()

    def stop(self):
        """Set the shutdown event for this thread and wait for it to stop."""
        if not self.started and not self.shutdown:
            return

        self._shutdown.set()
        self._stopped.wait()

    def run(self):
        """Threads main loop."""
        try:
            self._started.set()

            while True:
                self.callback()
                if self._shutdown.wait(self.interval):
                    break

        finally:
            self._stopped.set()


class PollingWatcher(Watcher):
    """Polls a directory recursively for changes.

    Detects file and directory changes, deletions, and creations. Recursion
    depth is limited to 2 levels. We use a limit because the scss file we're
    watching for changes could be sitting in the root of a project rather than
    a dedicated scss directory. That could lead to snapshots taking too long
    to build and diff. It's probably safe to assume that users aren't nesting
    scss deeper than a couple of levels.
    """

    def setup(self):
        """Set up the PollingWatcher.

        A PollingThread is created but not started.
        """
        self._snapshot_depth = 2
        self._snapshot = snapshots.take(self._watch_dir, self._snapshot_depth)
        self._thread = PollingThread(self.run, interval=1)

    def start(self):
        """Start the PollingThread."""
        self._thread.start()

    def stop(self):
        """Stop the PollingThread."""
        self._thread.stop()

    def join(self):
        """Wait for the PollingThread to finish.

        You should always call stop before join.
        """
        self._thread.join()

    def run(self):
        """Take a new snapshot and call on_change when a change is detected.

        Called repeatedly by the PollingThread.
        """
        next_snapshot = snapshots.take(self._watch_dir, self._snapshot_depth)
        changes = snapshots.diff(self._snapshot, next_snapshot)
        if changes:
            self._snapshot = next_snapshot
            self.on_change()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""The qtsass Watcher is responsible for watching and recompiling sass.

The default Watcher is the QtWatcher. If Qt is unavailable we fallback to the
PollingWatcher.
"""

# yapf: disable

from __future__ import absolute_import

# Local imports
from qtsass.watchers.polling import PollingWatcher


try:
    from qtsass.watchers.qt import QtWatcher
except ImportError:
    QtWatcher = None


# yapf: enable

Watcher = QtWatcher or PollingWatcher



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Libsass importers."""

# yapf: disable

from __future__ import absolute_import

# Standard library imports
import os

# Local imports
from qtsass.conformers import scss_conform


# yapf: enable


def norm_path(*parts):
    """Normalize path."""
    return os.path.normpath(os.path.join(*parts)).replace('\\', '/')


def qss_importer(*include_paths):
    """
    Return function which conforms imported qss files to valid scss.

    This fucntion is to be used as an importer for sass.compile.

    :param include_paths: Directorys containing scss, css, and sass files.
    """
    include_paths

    def find_file(import_file):
        # Create partial import filename
        dirname, basename = os.path.split(import_file)
        if dirname:
            import_partial_file = '/'.join([dirname, '_' + basename])
        else:
            import_partial_file = '_' + basename

        # Build potential file paths for @import "import_file"
        potential_files = []
        for ext in ['', '.scss', '.css', '.sass']:
            full_name = import_file + ext
            partial_name = import_partial_file + ext
            potential_files.append(full_name)
            potential_files.append(partial_name)
            for path in include_paths:
                potential_files.append(norm_path(path, full_name))
                potential_files.append(norm_path(path, partial_name))

        # Return first existing potential file
        for potential_file in potential_files:
            if os.path.isfile(potential_file):
                return potential_file

        return None

    def import_and_conform_file(import_file):
        """Return base file and conformed scss file."""
        real_import_file = find_file(import_file)
        with open(real_import_file, 'r') as f:
            import_str = f.read()

        return [(import_file, scss_conform(import_str))]

    return import_and_conform_file



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""qtsass command line interface."""

# yapf: disable

from __future__ import absolute_import, print_function

# Standard library imports
import argparse
import logging
import os
import sys
import time

# Local imports
from qtsass.api import (
    compile,
    compile_dirname,
    compile_filename,
    enable_logging,
    watch,
)


# yapf: enable

_log = logging.getLogger(__name__)


def create_parser():
    """Create qtsass's cli parser."""
    parser = argparse.ArgumentParser(
        prog='QtSASS',
        description='Compile a Qt compliant CSS file from a SASS stylesheet.',
    )
    parser.add_argument(
        'input',
        type=str,
        help='The SASS stylesheet file.',
    )
    parser.add_argument(
        '-o',
        '--output',
        type=str,
        help='The path of the generated Qt compliant CSS file.',
    )
    parser.add_argument(
        '-w',
        '--watch',
        action='store_true',
        help='If set, recompile when the source file changes.',
    )
    parser.add_argument(
        '-d',
        '--debug',
        action='store_true',
        help='Set the logging level to DEBUG.',
    )
    return parser


def main():
    """CLI entry point."""
    args = create_parser().parse_args()

    # Setup CLI logging
    debug = os.environ.get('QTSASS_DEBUG', args.debug)
    if debug in ('1', 'true', 'True', 'TRUE', 'on', 'On', 'ON', True):
        level = logging.DEBUG
    else:
        level = logging.INFO
    enable_logging(level)

    # Add a StreamHandler
    handler = logging.StreamHandler()
    if level == logging.DEBUG:
        fmt = '%(levelname)-8s: %(name)s> %(message)s'
        handler.setFormatter(logging.Formatter(fmt))
    logging.root.addHandler(handler)
    logging.root.setLevel(level)

    file_mode = os.path.isfile(args.input)
    dir_mode = os.path.isdir(args.input)

    if file_mode and not args.output:
        with open(args.input, 'r') as f:
            string = f.read()

        css = compile(
            string,
            include_paths=os.path.abspath(os.path.dirname(args.input)),
        )
        print(css)
        sys.exit(0)

    elif file_mode:
        _log.debug('compile_filename({}, {})'.format(args.input, args.output))
        compile_filename(args.input, args.output)

    elif dir_mode and not args.output:
        print('Error: missing required option: -o/--output')
        sys.exit(1)

    elif dir_mode:
        _log.debug('compile_dirname({}, {})'.format(args.input, args.output))
        compile_dirname(args.input, args.output)

    else:
        print('Error: input must be a file or a directory')
        sys.exit(1)

    if args.watch:
        _log.info('qtsass is watching {}...'.format(args.input))

        watcher = watch(args.input, args.output)
        watcher.start()
        try:
            while True:
                time.sleep(0.5)
        except KeyboardInterrupt:
            watcher.stop()
        watcher.join()
        sys.exit(0)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""qtsass - Compile SCSS files to valid Qt stylesheets."""

# yapf: disable

from __future__ import absolute_import, print_function

# Standard library imports
import logging
import os
import sys

# Third party imports
import sass

# Local imports
from qtsass.conformers import qt_conform, scss_conform
from qtsass.functions import qlineargradient, rgba
from qtsass.importers import qss_importer


if sys.version_info[0] == 3:
    from collections.abc import Mapping, Sequence
else:
    from collections import Mapping, Sequence


# yapf: enable

# Constants
DEFAULT_CUSTOM_FUNCTIONS = {'qlineargradient': qlineargradient, 'rgba': rgba}
DEFAULT_SOURCE_COMMENTS = False

# Logger setup
_log = logging.getLogger(__name__)


def compile(string, **kwargs):
    """
    Conform and Compile QtSASS source code to CSS.

    This function conforms QtSASS to valid SCSS before passing it to
    sass.compile. Any keyword arguments you provide will be combined with
    qtsass's default keyword arguments and passed to sass.compile.

    .. code-block:: python

        >>> import qtsass
        >>> qtsass.compile("QWidget {background: rgb(0, 0, 0);}")
        QWidget {background:black;}

    :param string: QtSASS source code to conform and compile.
    :param kwargs: Keyword arguments to pass to sass.compile
    :returns: CSS string
    """
    kwargs.setdefault('source_comments', DEFAULT_SOURCE_COMMENTS)
    kwargs.setdefault('custom_functions', [])
    kwargs.setdefault('importers', [])
    kwargs.setdefault('include_paths', [])

    # Add QtSass importers
    if isinstance(kwargs['importers'], Sequence):
        kwargs['importers'] = (list(kwargs['importers']) +
                               [(0, qss_importer(*kwargs['include_paths']))])
    else:
        raise ValueError('Expected Sequence for importers '
                         'got {}'.format(type(kwargs['importers'])))

    # Add QtSass custom_functions
    if isinstance(kwargs['custom_functions'], Sequence):
        kwargs['custom_functions'] = dict(
            DEFAULT_CUSTOM_FUNCTIONS,
            **{fn.__name__: fn
               for fn in kwargs['custom_functions']})
    elif isinstance(kwargs['custom_functions'], Mapping):
        kwargs['custom_functions'].update(DEFAULT_CUSTOM_FUNCTIONS)
    else:
        raise ValueError('Expected Sequence or Mapping for custom_functions '
                         'got {}'.format(type(kwargs['custom_functions'])))

    # Conform QtSass source code
    try:
        kwargs['string'] = scss_conform(string)
    except Exception:
        _log.error('Failed to conform source code')
        raise

    if _log.isEnabledFor(logging.DEBUG):
        from pprint import pformat
        log_kwargs = dict(kwargs)
        log_kwargs['string'] = 'Conformed SCSS<...>'
        _log.debug('Calling sass.compile with:')
        _log.debug(pformat(log_kwargs))
        _log.debug('Conformed scss:\n{}'.format(kwargs['string']))

    # Compile QtSass source code
    try:
        return qt_conform(sass.compile(**kwargs))
    except sass.CompileError:
        _log.error('Failed to compile source code')
        raise


def compile_filename(input_file, output_file, **kwargs):
    """Compile and save QtSASS file as CSS.

    .. code-block:: python

        >>> import qtsass
        >>> qtsass.compile_filename("dummy.scss", "dummy.css")

    :param input_file: Path to QtSass file.
    :param output_file: Path to write Qt compliant CSS.
    :param kwargs: Keyword arguments to pass to sass.compile
    :returns: CSS string
    """
    input_root = os.path.abspath(os.path.dirname(input_file))
    kwargs.setdefault('include_paths', [input_root])

    with open(input_file, 'r') as f:
        string = f.read()

    _log.info('Compiling {}...'.format(os.path.normpath(input_file)))
    css = compile(string, **kwargs)

    output_root = os.path.abspath(os.path.dirname(output_file))
    if not os.path.isdir(output_root):
        os.makedirs(output_root)

    with open(output_file, 'w') as css_file:
        css_file.write(css)
        _log.info('Created CSS file {}'.format(os.path.normpath(output_file)))

    return css


def compile_dirname(input_dir, output_dir, **kwargs):
    """Compiles QtSASS files in a directory including subdirectories.

    .. code-block:: python

        >>> import qtsass
        >>> qtsass.compile_dirname("./scss", "./css")

    :param input_dir: Directory containing QtSass files.
    :param output_dir: Directory to write compiled Qt compliant CSS files to.
    :param kwargs: Keyword arguments to pass to sass.compile
    """
    kwargs.setdefault('include_paths', [input_dir])

    def is_valid(file_name):
        return not file_name.startswith('_') and file_name.endswith('.scss')

    for root, _, files in os.walk(input_dir):
        relative_root = os.path.relpath(root, input_dir)
        output_root = os.path.join(output_dir, relative_root)
        fkwargs = dict(kwargs)
        fkwargs['include_paths'] = fkwargs['include_paths'] + [root]

        for file_name in [f for f in files if is_valid(f)]:
            scss_path = os.path.join(root, file_name)
            css_file = os.path.splitext(file_name)[0] + '.css'
            css_path = os.path.join(output_root, css_file)

            if not os.path.isdir(output_root):
                os.makedirs(output_root)

            compile_filename(scss_path, css_path, **fkwargs)


def enable_logging(level=None, handler=None):
    """Enable logging for qtsass.

    Sets the qtsass logger's level to:
        1. the provided logging level
        2. logging.DEBUG if the QTSASS_DEBUG envvar is a True value
        3. logging.WARNING

    .. code-block:: python
        >>> import logging
        >>> import qtsass
        >>> handler = logging.StreamHandler()
        >>> formatter = logging.Formatter('%(level)-8s: %(name)s> %(message)s')
        >>> handler.setFormatter(formatter)
        >>> qtsass.enable_logging(level=logging.DEBUG, handler=handler)

    :param level: Optional logging level
    :param handler: Optional handler to add
    """
    if level is None:
        debug = os.environ.get('QTSASS_DEBUG', False)
        if debug in ('1', 'true', 'True', 'TRUE', 'on', 'On', 'ON'):
            level = logging.DEBUG
        else:
            level = logging.WARNING

    logger = logging.getLogger('qtsass')
    logger.setLevel(level)
    if handler:
        logger.addHandler(handler)
    _log.debug('logging level set to {}.'.format(level))


def watch(source, destination, compiler=None, Watcher=None):
    """
    Watches a source file or directory, compiling QtSass files when modified.

    The compiler function defaults to compile_filename when source is a file
    and compile_dirname when source is a directory.

    :param source: Path to source QtSass file or directory.
    :param destination: Path to output css file or directory.
    :param compiler: Compile function (optional)
    :param Watcher: Defaults to qtsass.watchers.Watcher (optional)
    :returns: qtsass.watchers.Watcher instance
    """
    if os.path.isfile(source):
        watch_dir = os.path.dirname(source)
        compiler = compiler or compile_filename
    elif os.path.isdir(source):
        watch_dir = source
        compiler = compiler or compile_dirname
    else:
        raise ValueError('source arg must be a dirname or filename...')

    if Watcher is None:
        from qtsass.watchers import Watcher

    watcher = Watcher(watch_dir, compiler, (source, destination))
    return watcher



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Contains the Qt implementation of the Watcher api."""

# yapf: disable

from __future__ import absolute_import

# Local imports
from qtsass.watchers.polling import PollingWatcher


# We cascade through Qt bindings here rather than relying on a comprehensive
# Qt compatability library like qtpy or Qt.py. This prevents us from forcing a
# specific compatability library on users.
QT_BINDING = None
if not QT_BINDING:
    try:
        from PySide2.QtWidgets import QApplication
        from PySide2.QtCore import QObject, Signal
        QT_BINDING = 'pyside2'
    except ImportError:
        pass
if not QT_BINDING:
    try:
        from PyQt5.QtWidgets import QApplication
        from PyQt5.QtCore import QObject
        from PyQt5.QtCore import pyqtSignal as Signal
        QT_BINDING = 'pyqt5'
    except ImportError:
        pass
if not QT_BINDING:
    try:
        from PySide.QtGui import QApplication
        from PySide2.QtCore import QObject, Signal
        QT_BINDING == 'pyside'
    except ImportError:
        pass
if not QT_BINDING:
    from PyQt4.QtGui import QApplication
    from PyQt4.QtCore import QObject
    from PyQt4.QtCore import pyqtSignal as Signal
    QT_BINDING == 'pyqt4'


# yapf: enable


class QtDispatcher(QObject):
    """Used by QtWatcher to dispatch callbacks in the main ui thread."""

    signal = Signal()


class QtWatcher(PollingWatcher):
    """The Qt implementation of the Watcher api.

    Subclasses PollingWatcher but dispatches :meth:`compile_and_dispatch`
    using a Qt Signal to ensure that these calls are executed in the main ui
    thread. We aren't using a QFileSystemWatcher because it fails to report
    changes in certain circumstances.
    """

    _qt_binding = QT_BINDING

    def setup(self):
        """Set up QtWatcher."""
        super(QtWatcher, self).setup()
        self._qtdispatcher = None

    @property
    def qtdispatcher(self):
        """Get the QtDispatcher."""
        if self._qtdispatcher is None:
            self._qtdispatcher = QtDispatcher()
            self._qtdispatcher.signal.connect(self.compile_and_dispatch)
        return self._qtdispatcher

    def on_change(self):
        """Call when a change is detected."""
        self._log.debug('Change detected...')

        # If a QApplication event loop has not been started
        # call compile_and_dispatch in the current thread.
        if not QApplication.instance():
            return super(PollingWatcher, self).compile_and_dispatch()

        # Create and use a QtDispatcher to ensure compile and any
        # connected callbacks get executed in the main gui thread.
        self.qtdispatcher.signal.emit()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""qtsass command line interface."""

# yapf: disable

from __future__ import absolute_import

# Local imports
from qtsass import cli


# yapf: enable

if __name__ == '__main__':
    cli.main()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Libsass functions."""

# yapf: disable

# Third party imports
import sass


# yapf: enable


def rgba(r, g, b, a):
    """Convert r,g,b,a values to standard format.

    Where `a` is alpha! In CSS alpha can be given as:
     * float from 0.0 (fully transparent) to 1.0 (opaque)
    In Qt or qss that is:
     * int from 0 (fully transparent) to 255 (opaque)
    A percentage value 0% (fully transparent) to 100% (opaque) works
    in BOTH systems the same way!
    """
    result = 'rgba({}, {}, {}, {}%)'
    if isinstance(r, sass.SassNumber):
        if a.unit == '%':
            alpha = a.value
        elif a.value > 1.0:
            # A value from 0 to 255 is coming in, convert to %
            alpha = a.value / 2.55
        else:
            alpha = a.value * 100
        return result.format(
            int(r.value),
            int(g.value),
            int(b.value),
            int(alpha),
        )
    elif isinstance(r, float):
        return result.format(int(r), int(g), int(b), int(a * 100))


def rgba_from_color(color):
    """
    Conform rgba.

    :type color: sass.SassColor
    """
    # Inner rgba() call
    if not isinstance(color, sass.SassColor):
        return '{}'.format(color)

    return rgba(color.r, color.g, color.b, color.a)


def qlineargradient(x1, y1, x2, y2, stops):
    """
    Implement qss qlineargradient function for scss.

    :type x1: sass.SassNumber
    :type y1: sass.SassNumber
    :type x2: sass.SassNumber
    :type y2: sass.SassNumber
    :type stops: sass.SassList
    :return:
    """
    stops_str = []
    for stop in stops[0]:
        pos, color = stop[0]
        stops_str.append('stop: {} {}'.format(
            pos.value,
            rgba_from_color(color),
        ))
    template = 'qlineargradient(x1: {}, y1: {}, x2: {}, y2: {}, {})'
    return template.format(x1.value, y1.value, x2.value, y2.value,
                           ', '.join(stops_str))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Setup script for qtsass."""

# Standard library imports
from io import open
import ast
import os

# Third party imports
from setuptools import find_packages, setup


# Constants
HERE = os.path.abspath(os.path.dirname(__file__))


def get_version(module='qtsass'):
    """Get version."""
    with open(os.path.join(HERE, module, '__init__.py'), 'r') as f:
        data = f.read()

    lines = data.split('\n')
    for line in lines:
        if line.startswith('__version__'):
            version = ast.literal_eval(line.split('=')[-1].strip())
            break

    return version


def get_description():
    """Get long description."""
    with open(os.path.join(HERE, 'README.md'), 'r', encoding='utf-8') as f:
        data = f.read()

    return data


setup(
    name='qtsass310',
    version=get_version(),
    description='Compile SCSS files to valid Qt stylesheets.',
    long_description=get_description(),
    long_description_content_type='text/markdown',
    author='Yann Lanthony',
    maintainer='The Spyder Project Contributors',
    maintainer_email='qtsass@spyder-ide.org',
    url='https://github.com/spyder-ide/qtsass',
    license='MIT',
    packages=find_packages(exclude=['contrib', 'docs', 'tests*']),
    entry_points={
        'console_scripts': [
            'qtsass = qtsass.cli:main'
        ]
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    python_requires = ">=3.7",
    install_requires=[
        'libsass',
    ],
    keywords='qt sass qtsass scss css qss stylesheets',
)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015 Yann Lanthony
# Copyright (c) 2017-2018 Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (See LICENSE.txt for details)
# -----------------------------------------------------------------------------
"""Conform qss to compliant scss and css to valid qss."""

# yapf: disable

from __future__ import absolute_import, print_function

# Standard library imports
import re


# yapf: enable

_DEFAULT_COORDS = ('x1', 'y1', 'x2', 'y2')


class Conformer(object):
    """Base class for all text transformations."""

    def to_scss(self, qss):
        """Transform some qss to valid scss."""
        return NotImplemented

    def to_qss(self, css):
        """Transform some css to valid qss."""
        return NotImplemented


class NotConformer(Conformer):
    """Conform QSS "!" in selectors."""

    def to_scss(self, qss):
        """Replace "!" in selectors with "_qnot_"."""
        return qss.replace(':!', ':_qnot_')

    def to_qss(self, css):
        """Replace "_qnot_" in selectors with "!"."""
        return css.replace(':_qnot_', ':!')


class QLinearGradientConformer(Conformer):
    """Conform QSS qlineargradient function."""

    qss_pattern = re.compile(
        r'qlineargradient\('
        r'((?:(?:\s+)?(?:x1|y1|x2|y2):(?:\s+)?[0-9A-Za-z$_\.-]+,?)+)'  # coords
        r'((?:(?:\s+)?stop:.*,?)+(?:\s+)?)?'  # stops
        r'\)',
        re.MULTILINE,
    )

    def _conform_coords_to_scss(self, group):
        """
        Take a qss str with xy coords and returns the values.

          'x1: 0, y1: 0, x2: 0, y2: 0' => '0, 0, 0, 0'
          'y1: 1' => '0, 1, 0, 0'
        """
        values = ['0', '0', '0', '0']
        for key_values in [part.split(':', 1) for part in group.split(',')]:
            try:
                key, value = key_values
                key = key.strip()
                if key in _DEFAULT_COORDS:
                    pos = _DEFAULT_COORDS.index(key)
                    if pos >= 0 and pos <= 3:
                        values[pos] = value.strip()
            except ValueError:
                pass
        return ', '.join(values)

    def _conform_stops_to_scss(self, group):
        """
        Take a qss str with stops and returns the values.

          'stop: 0 red, stop: 1 blue' => '0 red, 1 blue'
        """
        new_group = []
        split = [""]
        bracket_level = 0
        for char in group:
            if not bracket_level and char == ",":
                split.append("")
                continue
            elif char == "(":
                bracket_level += 1
            elif char == ")":
                bracket_level -= 1
            split[-1] += char

        for part in split:
            if part:
                _, value = part.split(':', 1)
                new_group.append(value.strip())
        return ', '.join(new_group)

    def to_scss(self, qss):
        """
        Conform qss qlineargradient to scss qlineargradient form.

        Normalize all whitespace including the removal of newline chars.

        qlineargradient(x1: 0, y1: 0, x2: 0, y2: 0, stop: 0 red, stop: 1 blue)
        =>
        qlineargradient(0, 0, 0, 0, (0 red, 1 blue))
        """
        conformed = qss

        for coords, stops in self.qss_pattern.findall(qss):
            new_coords = self._conform_coords_to_scss(coords)
            conformed = conformed.replace(coords, new_coords, 1)

            if not stops:
                continue

            new_stops = ', ({})'.format(self._conform_stops_to_scss(stops))
            conformed = conformed.replace(stops, new_stops, 1)

        return conformed

    def to_qss(self, css):
        """Transform to qss from css."""
        return css


conformers = [c() for c in Conformer.__subclasses__() if c is not Conformer]


def scss_conform(input_str):
    """
    Conform qss to valid scss.

    Runs the to_scss method of all Conformer subclasses on the input_str.
    Conformers are run in order of definition.

    :param input_str: QSS string
    :returns: Valid SCSS string
    """
    conformed = input_str
    for conformer in conformers:
        conformed = conformer.to_scss(conformed)

    return conformed


def qt_conform(input_str):
    """
    Conform css to valid qss.

    Runs the to_qss method of all Conformer subclasses on the input_str.
    Conformers are run in reverse order.

    :param input_str: CSS string
    :returns: Valid QSS string
    """
    conformed = input_str
    for conformer in conformers[::-1]:
        conformed = conformer.to_qss(conformed)

    return conformed
